"""
GUI framework and application for use with Python unit testing framework.
Executes test cases and suites written using the framework provided by the
'unittest' module, included with Python.

Further information is available in the documentation, and from

 http://projects.edgewall.com/qunittest/

Copyright (C) 2004 Edgewall Software
Copyright (C) 2004 Daniel Lundin <daniel@edgewall.com>
"""

from __init__ import __version__
from __init__ import __credits__

import imp
import sys
import os
import os.path
import time
import types
import unittest
import traceback

import form_main
import form_about

from qt import *


class QtTestResult(unittest.TestResult):
    def __init__(self, callback):
        unittest.TestResult.__init__(self)
        self.callback = callback

    def addError(self, test, err):
        unittest.TestResult.addError(self, test, err)
        self.callback.notifyTestError(test, err)

    def addSuccess(self, test):
        unittest.TestResult.addSuccess(self, test)
        self.callback.notifyTestSuccess(test)

    def addFailure(self, test, err):
        unittest.TestResult.addFailure(self, test, err)
        self.callback.notifyTestFailed(test, err)

    def stopTest(self, test):
        unittest.TestResult.stopTest(self, test)
        self.callback.notifyTestFinished(test)

    def startTest(self, test):
        unittest.TestResult.startTest(self, test)
        self.callback.notifyTestStarted(test)


class RollbackImporter:
    def __init__(self):
        self.previousModules = sys.modules.copy()

    def rollbackImports(self):
        for modname in sys.modules.keys():
            if not self.previousModules.has_key(modname):
                del(sys.modules[modname])


class TestListItem(QListBoxText):

    def __init__(self, txt):
        QListBoxText.__init__(self)
        self.setText(txt)
        self.setStatus(0)
        self._bgerr = QBrush(QColor(200,0,0))
        self._fgerr = QColor(255,255,180)

    def setStatus(self, status=0):
        self._status = status

    def paint(self, painter):
        if not self.isSelected():
            if self._status:
                painter.fillRect(painter.viewport(), self._bgerr);
                if self._status == 1:
                    painter.setPen(Qt.white)
                else:
                    painter.setPen(self._fgerr)
        return QListBoxText.paint(self,painter)


class FakeStdout:
    def __init__(self, textedit):
        self._textedit = textedit

    def write(self, text):
        self._textedit.append(text)
        self._textedit.moveCursor(QTextEdit.MoveEnd, 0)

    def flush(self):
        pass

    def isatty(self):
        return 1

class AboutWindow(form_about.Form2):

    _about_text = """
<p align="center">
<font size="+3" color="#bb0000"><b>QTestRunner</b></font> v%(ver)s<br>
Copyright &copy; 2004, <a href="http://www.edgewall.com./">Edgewall software</a>
<br>
<br>
By: Daniel Lundin &lt;<a href="mailto:daniel@edgewall.com">daniel@edgewall.com</a>&gt;
</p>
""" % {'ver':__version__}

    def __init__(self):
        form_about.Form2.__init__(self,None,None,0,0)
        self.textLabel6.setText(self._about_text)
        self.setCaption('About QUnitTest %s' % __version__)

class MainWindow(form_main.Form1):

    def __init__(self, app, fname=None, run=0):
        form_main.Form1.__init__(self,parent = None,name = None,fl = 0)
        self._app = app
        self._rollback_importer = None
        self._about = AboutWindow()
        self.setCaption('QUnitTest %s' % __version__)
        pal = self.progressBar1.palette().copy()
        self.progressBar1.setPalette(pal)
        self._sys_path = sys.path
        self.loader = unittest.TestLoader()
        self.stdout = FakeStdout(self.outputTextEdit)
        self.resetEnv()
        self.enableRun(enable=0)
        if fname:
            self.fileOpen(fname)
        if run:
            self.testRun()

    def resetEnv(self):
        self._test_name = None
        self._running = 0
        self._numtests = 0
        self._cases = {}
        self._suites = {}
        self._module = None
        self._filename = None
        self.resetRun()

    def resetRun(self):
        self._run_count = 0
        self._err_count = 0
        self._fail_count = 0
        self.setRunCount(0)
        self.setFailCount(0)
        self.setErrCount(0)
        self.setRemCount(self._numtests)
        self.resultTextEdit1.setText('')
        self.outputTextEdit.setText('')
        self.setProgress(0)
        self.setProgressBarStatus(0)

    def errorMessage(self, title, msg, show_tb=0):
        if show_tb:
            lt, lv, ltb = sys.exc_info()
            tbmsg = ''.join(traceback.format_exception(lt, lv, ltb))
            msg = "%s\n\n%s" % (msg, tbmsg)
        QMessageBox.warning(None, title, msg)

    def enableRun(self, enable=1):
        self.runButton1.setEnabled(enable)
        self.comboTestList.setEnabled(enable)
        self.testRunAction.setEnabled(enable)
        self.progressBar1.setEnabled(enable)

    def setProgressBarStatus(self, status=0):
        if status == 1:
            c = Qt.green
        elif status == 2:
            c = Qt.red
        else:
            c = self.palette().color(QPalette.Active, QColorGroup.Highlight)
        self.progressBar1.palette().setColor(QColorGroup.Highlight, c)
        self.progressBar1.update()

    def fileExit(self):
        self._app.quit()

    def helpAbout(self):
        self._about.show()

    def fileOpen(self, fname=None):
        if not fname:
            fname = QFileDialog.getOpenFileName(None,"*.py",self)
            if fname.isEmpty():
                self.setStatusBar('Loading aborted')
                return
            fname = str(fname)
        if self.loadTests(fname, 1):
            self.setStatusBar("Loaded tests from '%s'" % fname)
            self.enableRun()
            self.setCaption('QUnitTest %s - %s' % (__version__,
                                                   os.path.basename(fname)))


    def loadTests(self, fname,update_env=0):
        if self._rollback_importer:
            self._rollback_importer.rollbackImports()
        self._rollback_importer = RollbackImporter()

        dir_name = os.path.dirname(fname)
        mod_name = os.path.basename(fname)[:-3]
        try:
            f = file(fname, "r")
            sys.path = [dir_name] + self._sys_path
            mod = imp.load_source(mod_name, fname, f)
            f.close()
            reload(mod)
        except Exception, e:
            self.errorMessage("Error loading module",
                              "Error loading module '%s' from '%s'" %
                              (mod_name, fname), 1)
            return 0
        tests = {}
        for name in dir(mod):
            obj = getattr(mod, name)
            if type(obj) == type or type(obj) == types.ClassType:
                if issubclass(obj, unittest.TestCase) \
                       or issubclass(obj, unittest.TestSuite):
                    tests[name] = obj
            elif type(obj) == types.FunctionType and name == 'suite':
                tests['suite'] = obj

        if not tests:
            QMessageBox.information(None, "No tests found",
                     "No test suites or cases found in module '%s'" % mod_name)
            return 0
        tn = self._test_name
        self.resetEnv()
        self._filename = fname
        self._module = mod
        self._tests = tests
        if update_env:
            self.comboTestList.clear()
            self.comboTestList.insertStrList(tests.keys())
        self.testChange(tn)
        return 1

    def testChange(self, tname=None):

        def _get_suite_tests(suite):
            tests = []
            for tst in suite._tests:
                if issubclass(tst.__class__, unittest.TestSuite):
                    tests += _get_suite_tests(tst)
                else:
                    t_name = tst._TestCase__testMethodName
                    cls = tst.__class__
                    t_repr = '%s.%s.%s' % \
                    (cls.__module__, cls.__name__,
                     tst._TestCase__testMethodName)
                    tests.append(t_repr)
            return tests
        tname = str(tname)
        try:
            test_class = self._tests[tname]
        except KeyError:
            return self.testChange(self._tests.keys()[0])
        self._test_name = tname
        self._test_is_suite = 0
        self.resetRun()
        self.resultListBox1.clear()
        self._list_items = {}

        if type(test_class) == types.FunctionType \
               or issubclass(test_class, unittest.TestSuite):
            suite = test_class()
            tests = _get_suite_tests (suite)
            self._test_is_suite = 1
        else:
            tests = filter(lambda s: s[:4] == 'test', dir(test_class))
            self._test_is_suite = 0

        for t in tests:
            it = TestListItem(t)
            self._list_items[t] = it
            self.resultListBox1.insertItem(it)

        self._numtests = len(tests)
        self.setRemCount(self._numtests)
        self.progressBar1.setTotalSteps(self._numtests)

    def setRunCount(self, count):
        self.runsLabel.setText(
            '<font color="#00007f"><b>%i</b></font>' % count)

    def setFailCount(self, count):
        self.failLabel.setText(
            '<font color="#f08200"><b>%i</b></font>' % count)

    def setErrCount(self, count):
        self.errorLabel.setText(
            '<font color="#bb0000"><b>%i</b></font>' % count)

    def setRemCount(self, count):
        self.remainingLabel.setText(
            '<font color="#00007f"><b>%i</b></font>' % count)

    def setProgress(self, prog):
        self.progressBar1.setProgress(prog)

    def setStatusBar(self, msg, delay=2000):
        self.statusBar().message(msg, delay)

    def setTestListItemStatus(self, idx, status):
        li = self.resultListBox1.item(idx - 1)
        if li:
            li.setStatus(status)
            self.resultListBox1.viewport().update()

    def testRun(self):
        self.resetRun()
        self.setStatusBar('Running test')
        try:
            if self._filename:
                self.loadTests(self._filename)

            test_name = str(self.comboTestList.currentText())
            if test_name[-2] == '()':
                test_name = test_name[:-2]
            obj = self._tests[test_name]
            test = self.loader.loadTestsFromName(test_name, self._module)
            result = QtTestResult(self)
            so = sys.stdout
            sys.stdout = self.stdout
            startTime = time.time()
            test.run(result)
            stopTime = time.time()
            sys.stdout = so
            timeTaken = float(stopTime - startTime)

            msg = "Ran %i tests in %0.4fs" % (self._run_count, timeTaken)
            self.setStatusBar(msg)
            msg = "<i>%s</i>" % msg
            if self._fail_count == 0 and self._err_count == 0:
                msg += '<hr><font size="+2" color="#00bb00"><b>All OK</b></font>'
                self.setProgressBarStatus(1)
            else:
                self.setProgressBarStatus(2)
            self.logResultMessage(msg)

        except Exception, e:
            self.errorMessage("Error running test",
                              "An error occured while running the tests", 1)

    def logResultMessage(self, msg):
            self.resultTextEdit1.append(msg)

    def logResultTest(self, test, err=None, is_error=0):
        msg = '<b>%s</b> ... ' % test._TestCase__testMethodName
        if err:
            if is_error:
                msg += '<font color="#fb0000"><b>Error!</b></font>'
            else:
                msg += '<font color="#fbb200"><b>Failed!</b></font>'
            lt, lv, ltb = err
            tb = ''.join(traceback.format_exception(*err))
            tb = tb.replace(' ','&nbsp;')
            tb = tb.replace('\n','<br>')
            msg += '<br>' + tb
        else:
            msg += '<font color="#00bb00"><b>OK</b></font>'
        msg += '<hr>'
        self.logResultMessage(msg)


    def notifyTestStarted(self, test):
        self._run_count += 1
        self.setRunCount(self._run_count)
        self.setRemCount(self._numtests - self._run_count)
        self.setProgress(self._run_count)

    def notifyTestFinished(self, test):
        pass

    def notifyTestSuccess(self, test):
        self.logResultTest(test)

    def notifyTestFailed(self, test, err):
        self._fail_count += 1
        self.setFailCount(self._fail_count)
        self.setTestListItemStatus(self._run_count, 2)
        self.logResultTest(test, err)

    def notifyTestError(self, test, err):
        self._err_count += 1
        self.setErrCount(self._err_count)
        self.setTestListItemStatus(self._run_count, 1)
        self.logResultTest(test, err, 1)


_USAGE = """\
Usage: %(progName)s [options] [testfile.py]

Options:
  -h, --help       Show this message
  -v, --version    Show version information

Examples:
  %(progName)s                               - Start QUnitTest
  %(progName)s my_tests.py                   - Load test suite[s] on start

%(credits)s
""" 

def main():
    fn = None
    argc = len(sys.argv)
    if argc > 1:
        if '-h' in sys.argv or '--help' in sys.argv:
            print _USAGE % {'progName':sys.argv[0], 'credits':__credits__}
            sys.exit(0)
        elif '-v' in sys.argv or '--version' in sys.argv:
            print 'QUnitTest %s\n\n%s' % (__version__, __credits__)
            sys.exit(0)
        else:
            fn = sys.argv[-1]
    run = (argc == 3 and ('-r' in sys.argv or '--run' in sys.argv))

    a = QApplication(sys.argv)
    mw = MainWindow(a, fn, run)
    a.connect(a, SIGNAL('lastWindowClosed()'), a, SLOT('quit()'))
    mw.show()
    a.exec_loop()

if __name__ == '__main__':
    main()
